#!/usr/bin/env python
#Author: Mark Feineigle
#Create Date: 2018/03/15
#Last Modify: 2018/03/15
# Mastering Machine Learning with Python in Six Steps page 384

# page 384
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from sklearn.neural_network import MLPClassifier
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from sklearn.metrics import confusion_matrix
from sklearn.datasets import load_digits
np.random.seed(seed=2017)
# load data
digits = load_digits()
print('We have %d samples'%len(digits.target))
## plot the first 32 samples to get a sense of the data
fig = plt.figure(figsize = (8,8))
fig.subplots_adjust(left=0, right=1, bottom=0, top=1, hspace=0.05, wspace=0.05)
for i in range(32):
  ax = fig.add_subplot(8, 8, i+1, xticks=[], yticks=[])
  ax.imshow(digits.images[i], cmap=plt.cm.gray_r)
  ax.text(0, 1, str(digits.target[i]), bbox=dict(facecolor='white'))
plt.show()


# split data to training and testing data
X_train, X_test, y_train, y_test = train_test_split(digits.data,
                                                    digits.target,
                                                    test_size=0.2,
                                                    random_state=2017)
print 'Number of samples in training set: %d' % (len(y_train))
print 'Number of samples in test set: %d' % (len(y_test))
# Standardise data, and fit only to the training data
scaler = StandardScaler()
scaler.fit(X_train)
# Apply the transformations to the data
X_train_scaled = scaler.transform(X_train)
X_test_scaled = scaler.transform(X_test)
# Initialize ANN classifier
mlp = MLPClassifier(hidden_layer_sizes=(100),
                    activation='logistic',
                    max_iter=100)
# Train the classifier with the training data
mlp.fit(X_train_scaled, y_train)


# page 386
#MLPClassifier(activation='logistic', #NOTE this isn't doing anything?
#              alpha=0.0001,
#              batch_size='auto',
#              beta_1=0.9,
#              beta_2=0.999,
#              early_stopping=False,
#              epsilon=1e-08,
#              hidden_layer_sizes=(30, 30, 30),
#              learning_rate='constant',
#              learning_rate_init=0.001,
#              max_iter=100,
#              momentum=0.9,
#              nesterovs_momentum=True,
#              power_t=0.5,
#              random_state=None,
#              shuffle=True,
#              solver='adam',
#              tol=0.0001,
#              validation_fraction=0.1,
#              verbose=False,
#              warm_start=False)
print("Training set score: %f" % mlp.score(X_train_scaled, y_train))
print("Test set score: %f" % mlp.score(X_test_scaled, y_test))


# page 387
# predict results from the test data
X_test_predicted = mlp.predict(X_test_scaled)
fig = plt.figure(figsize=(8, 8)) # figure size in inches
fig.subplots_adjust(left=0, right=1, bottom=0, top=1, hspace=0.05, wspace=0.05)
# plot the digits: each image is 8x8 pixels
for i in range(32):
  ax = fig.add_subplot(8, 8, i + 1, xticks=[], yticks=[])
  ax.imshow(X_test.reshape(-1, 8, 8)[i], cmap=plt.cm.gray_r)
  # label the image with the target value
  if X_test_predicted[i] == y_test[i]:
    ax.text(0, 1, X_test_predicted[i], color='green', bbox=dict(facecolor='white'))
  else:
    ax.text(0, 1, X_test_predicted[i], color='red', bbox=dict(facecolor='white'))
plt.show()


# page 389
# Function to nudge the dataset
def nudge_dataset(X, Y):
  """
  This produces a dataset 5 times bigger than the
  original one,
  by moving the 8x8 images in X around by 1px to left,
  right, down, up
  """
  direction_vectors = [[[0, 1, 0],
                       [0, 0, 0],
                       [0, 0, 0]],
                       [[0, 0, 0],
                       [1, 0, 0],
                       [0, 0, 0]],
                       [[0, 0, 0],
                       [0, 0, 1],
                       [0, 0, 0]],
                       [[0, 0, 0],
                       [0, 0, 0],
                       [0, 1, 0]]]
  shift = lambda x, w: convolve(x.reshape((8, 8)), mode='constant', weights=w).ravel()
  X = np.concatenate([X] + [np.apply_along_axis(shift, 1, X, vector)for vector in direction_vectors])
  Y = np.concatenate([Y for _ in range(5)], axis=0)
  return X, Y


# page 390
# Example adapted from scikit-learn documentation
import numpy as np
import matplotlib.pyplot as plt
from sklearn import linear_model, datasets, metrics
from sklearn.model_selection import train_test_split
from sklearn.neural_network import BernoulliRBM
from sklearn.pipeline import Pipeline
from scipy.ndimage import convolve
from sklearn.grid_search import GridSearchCV
from sklearn.linear_model import LogisticRegression
# Load Data
digits = datasets.load_digits()
X = np.asarray(digits.data, 'float32')
y = digits.target
X, y = nudge_dataset(X, digits.target)
# Scale the features such that the values are between 0-1 scale
X = (X - np.min(X, 0)) / (np.max(X, 0) + 0.0001)
X_train, X_test, y_train, y_test = train_test_split(X, y,
test_size=0.2, random_state=2017)
print X.shape
print y.shape


# page 391
# Gridsearch for logistic regression
# perform a grid search on the 'C' parameter of Logistic
params = {"C": [1.0, 10.0, 100.0]}
Grid_Search = GridSearchCV(LogisticRegression(), params, n_jobs = -1, verbose = 1)
Grid_Search.fit(X_train, y_train)
# print diagnostic information to the user and grab the
print "Best Score: %0.3f" % (Grid_Search.best_score_)
# best model
bestParams = Grid_Search.best_estimator_.get_params()
print bestParams.items()


# page 392
# evaluate using Logistic Regression and only the raw pixel
logistic = LogisticRegression(C = 100)
logistic.fit(X_train, y_train)
print "Train accuracy: ", metrics.accuracy_score(y_train, logistic.predict(X_train))
print "Test accuracyL ", metrics.accuracy_score(y_test, logistic.predict(X_test))


# page 392
# initialize the RBM + Logistic Regression pipeline
rbm = BernoulliRBM()
logistic = LogisticRegression()
classifier = Pipeline([("rbm", rbm), ("logistic", logistic)])
params = {"rbm__learning_rate": [0.1, 0.01, 0.001],
          "rbm__n_iter": [20, 40, 80],
          "rbm__n_components": [50, 100, 200],
          "logistic__C": [1.0, 10.0, 100.0]}
# perform a grid search over the parameter
Grid_Search = GridSearchCV(classifier, params, n_jobs = -1, verbose = 1)
Grid_Search.fit(X_train, y_train)
# print diagnostic information to the user and grab the
# best model
print "Best Score: %0.3f" % (Grid_Search.best_score_)
print "RBM + Logistic Regression parameters"
bestParams = Grid_Search.best_estimator_.get_params()
# loop over the parameters and print each of them out
# so they can be manually set
for p in sorted(params.keys()):
  print "\t %s: %f" % (p, bestParams[p])


# initialize the RBM + Logistic Regression classifier with
# the cross-validated parameters
rbm = BernoulliRBM(n_components = 200,
                   n_iter = 20,
                   learning_rate = 0.1,
                   verbose = False)
                   logistic = LogisticRegression(C = 100)
# train the classifier and show an evaluation report
classifier = Pipeline([("rbm", rbm), ("logistic", logistic)])
classifier.fit(X_train, y_train)
print metrics.accuracy_score(y_train, classifier.predict(X_train))
print metrics.accuracy_score(y_test, classifier.predict(X_test))

# plot RBM components
plt.figure(figsize=(15, 15))
for i, comp in enumerate(rbm.components_):
  plt.subplot(20, 20, i + 1)
  plt.imshow(comp.reshape((8, 8)), cmap=plt.cm.gray_r, interpolation='nearest')
  plt.xticks(())
  plt.yticks(())
plt.suptitle('200 components extracted by RBM', fontsize=16)
plt.show()

